==> bucketlib-fabric-1.21.3-4.2.4.0-sources.jar: de/cech12/bucketlib/api/crafting/EmptyIngredient.java <==
package de.cech12.bucketlib.api.crafting;

import com.mojang.serialization.MapCodec;
import com.mojang.serialization.codecs.RecordCodecBuilder;
import de.cech12.bucketlib.BucketLibMod;
import de.cech12.bucketlib.api.BucketLib;
import de.cech12.bucketlib.api.item.UniversalBucketItem;
import de.cech12.bucketlib.util.BucketLibUtil;
import net.fabricmc.fabric.api.recipe.v1.ingredient.CustomIngredient;
import net.fabricmc.fabric.api.recipe.v1.ingredient.CustomIngredientSerializer;
import net.minecraft.class_1792;
import net.minecraft.class_1799;
import net.minecraft.class_1802;
import net.minecraft.class_2960;
import net.minecraft.class_6862;
import net.minecraft.class_6880;
import net.minecraft.class_7923;
import net.minecraft.class_7924;
import net.minecraft.class_9129;
import net.minecraft.class_9139;
import javax.annotation.Nonnull;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.Optional;

public class EmptyIngredient implements CustomIngredient {

    protected class_1792 item;
    protected class_6862<class_1792> tag;
    private List<class_6880<class_1792>> matchingStacks;

    public EmptyIngredient(class_1792 item, class_6862<class_1792> tag) {
        this.item = item;
        this.tag = tag;
    }

    public EmptyIngredient(class_1792 item) {
        this(item, null);
    }

    public EmptyIngredient(class_6862<class_1792> tag) {
        this(null, tag);
    }

    public EmptyIngredient() {
        this((class_1792) null, null);
    }

    public EmptyIngredient(Optional<class_2960> itemOptional, Optional<class_6862<class_1792>> tagOptional) {
        this(itemOptional.map(class_7923.field_41178::method_10223).filter(Optional::isPresent).map(reference -> reference.get().comp_349()).orElse(null), tagOptional.orElse(null));
    }

    @Override
    public boolean test(class_1799 itemStack) {
        if (itemStack == null || itemStack.method_7960()) {
            return false;
        }
        if (this.item == null && this.tag == null && itemStack.method_7909() == class_1802.field_8550) {
            return true;
        }
        if (this.item != null && itemStack.method_7909() == this.item
                || this.tag != null && itemStack.method_31573(this.tag)
                || this.item == null && this.tag == null && itemStack.method_7909() instanceof UniversalBucketItem) {
            return BucketLibUtil.isEmpty(itemStack);
        }
        return false;
    }

    @Override
    public List<class_6880<class_1792>> getMatchingItems() {
        if (this.matchingStacks == null) {
            this.matchingStacks = new ArrayList<>();
            if (this.item == null && this.tag == null) {
                this.matchingStacks.add(class_6880.method_40223(class_1802.field_8550));
            }
            BucketLibMod.getRegisteredBuckets().forEach(universalBucketItem -> {
                class_1799 universalBucketItemStack = new class_1799(universalBucketItem);
                if (this.item != null && universalBucketItem == this.item
                        || this.tag != null && universalBucketItemStack.method_31573(this.tag)
                        || this.item == null && this.tag == null) {
                    this.matchingStacks.add(class_6880.method_40223(universalBucketItem));

                }
            });
        }
        return this.matchingStacks;
    }

    @Override
    public boolean requiresTesting() {
        return true;
    }

    @Override
    public CustomIngredientSerializer<?> getSerializer() {
        return Serializer.INSTANCE;
    }

    public static final class Serializer implements CustomIngredientSerializer<EmptyIngredient> {

        public static final Serializer INSTANCE = new Serializer();
        public static final class_2960 NAME = BucketLib.id("empty");

        private static final MapCodec<EmptyIngredient> CODEC = RecordCodecBuilder.mapCodec(builder ->
                builder.group(
                        class_2960.field_25139.optionalFieldOf("item").forGetter(i -> Optional.of(class_7923.field_41178.method_10221(i.item))),
                        class_6862.method_40090(class_7923.field_41178.method_46765()).optionalFieldOf("tag").forGetter(i -> Optional.ofNullable(i.tag))
                ).apply(builder, EmptyIngredient::new)
        );

        private static final class_9139<class_9129, EmptyIngredient> PACKET_CODEC = class_9139.method_56437(
                EmptyIngredient.Serializer::write,
                EmptyIngredient.Serializer::read);

        private Serializer() {}

        @Override
        public class_2960 getIdentifier() {
            return NAME;
        }

        @Override
        public MapCodec<EmptyIngredient> getCodec() {
            return CODEC;
        }

        @Override
        public class_9139<class_9129, EmptyIngredient> getPacketCodec() {
            return PACKET_CODEC;
        }

        @Nonnull
        private static EmptyIngredient read(class_9129 buffer) {
            String item = buffer.method_19772();
            String tagId = buffer.method_19772();
            if (!item.isEmpty()) {
                Optional<class_6880.class_6883<class_1792>> itemOptional = class_7923.field_41178.method_10223(class_2960.method_60654(item));
                if (itemOptional.isPresent()) {
                    return new EmptyIngredient(itemOptional.get().comp_349());
                }
            }
            if (!tagId.isEmpty()) {
                class_6862<class_1792> tag = class_6862.method_40092(class_7924.field_41197, class_2960.method_60654(tagId));
                return new EmptyIngredient(tag);
            }
            return new EmptyIngredient();
        }

        private static void write(@Nonnull class_9129 buffer, @Nonnull EmptyIngredient ingredient) {
            buffer.method_10814(ingredient.item != null ? Objects.requireNonNull(class_7923.field_41178.method_10221(ingredient.item)).toString() : "");
            buffer.method_10814(ingredient.tag != null ? ingredient.tag.comp_327().toString() : "");
        }
    }

}

==> bucketlib-fabric-1.21-4.1.6.0-sources.jar: de/cech12/bucketlib/api/crafting/EmptyIngredient.java <==
package de.cech12.bucketlib.api.crafting;

import com.mojang.serialization.MapCodec;
import com.mojang.serialization.codecs.RecordCodecBuilder;
import de.cech12.bucketlib.BucketLibMod;
import de.cech12.bucketlib.api.BucketLib;
import de.cech12.bucketlib.api.item.UniversalBucketItem;
import de.cech12.bucketlib.util.BucketLibUtil;
import net.fabricmc.fabric.api.recipe.v1.ingredient.CustomIngredient;
import net.fabricmc.fabric.api.recipe.v1.ingredient.CustomIngredientSerializer;
import net.minecraft.class_1792;
import net.minecraft.class_1799;
import net.minecraft.class_1802;
import net.minecraft.class_2960;
import net.minecraft.class_6862;
import net.minecraft.class_7923;
import net.minecraft.class_7924;
import net.minecraft.class_9129;
import net.minecraft.class_9139;
import javax.annotation.Nonnull;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.Optional;

public class EmptyIngredient implements CustomIngredient {

    protected class_1792 item;
    protected class_6862<class_1792> tag;
    private List<class_1799> matchingStacks;

    public EmptyIngredient(class_1792 item, class_6862<class_1792> tag) {
        this.item = item;
        this.tag = tag;
    }

    public EmptyIngredient(class_1792 item) {
        this(item, null);
    }

    public EmptyIngredient(class_6862<class_1792> tag) {
        this(null, tag);
    }

    public EmptyIngredient() {
        this((class_1792) null, null);
    }

    public EmptyIngredient(Optional<class_2960> itemOptional, Optional<class_6862<class_1792>> tagOptional) {
        this(itemOptional.map(class_7923.field_41178::method_10223).orElse(null), tagOptional.orElse(null));
    }

    @Override
    public boolean test(class_1799 itemStack) {
        if (itemStack == null || itemStack.method_7960()) {
            return false;
        }
        if (this.item == null && this.tag == null && itemStack.method_7909() == class_1802.field_8550) {
            return true;
        }
        if (this.item != null && itemStack.method_7909() == this.item
                || this.tag != null && itemStack.method_31573(this.tag)
                || this.item == null && this.tag == null && itemStack.method_7909() instanceof UniversalBucketItem) {
            return BucketLibUtil.isEmpty(itemStack);
        }
        return false;
    }

    @Override
    public List<class_1799> getMatchingStacks() {
        if (this.matchingStacks == null) {
            this.matchingStacks = new ArrayList<>();
            if (this.item == null && this.tag == null) {
                this.matchingStacks.add(new class_1799(class_1802.field_8550));
            }
            BucketLibMod.getRegisteredBuckets().forEach(universalBucketItem -> {
                class_1799 universalBucketItemStack = new class_1799(universalBucketItem);
                if (this.item != null && universalBucketItem == this.item
                        || this.tag != null && universalBucketItemStack.method_31573(this.tag)
                        || this.item == null && this.tag == null) {
                    this.matchingStacks.add(universalBucketItemStack);
                }
            });
        }
        return this.matchingStacks;
    }

    @Override
    public boolean requiresTesting() {
        return true;
    }

    @Override
    public CustomIngredientSerializer<?> getSerializer() {
        return Serializer.INSTANCE;
    }

    public static final class Serializer implements CustomIngredientSerializer<EmptyIngredient> {

        public static final Serializer INSTANCE = new Serializer();
        public static final class_2960 NAME = BucketLib.id("empty");

        private static final MapCodec<EmptyIngredient> CODEC = RecordCodecBuilder.mapCodec(builder ->
                builder.group(
                        class_2960.field_25139.optionalFieldOf("item").forGetter(i -> Optional.of(class_7923.field_41178.method_10221(i.item))),
                        class_6862.method_40090(class_7923.field_41178.method_30517()).optionalFieldOf("tag").forGetter(i -> Optional.ofNullable(i.tag))
                ).apply(builder, EmptyIngredient::new)
        );

        private static final class_9139<class_9129, EmptyIngredient> PACKET_CODEC = class_9139.method_56437(
                EmptyIngredient.Serializer::write,
                EmptyIngredient.Serializer::read);

        private Serializer() {}

        @Override
        public class_2960 getIdentifier() {
            return NAME;
        }

        @Override
        public MapCodec<EmptyIngredient> getCodec(boolean allowEmpty) {
            return CODEC;
        }

        @Override
        public class_9139<class_9129, EmptyIngredient> getPacketCodec() {
            return PACKET_CODEC;
        }

        @Nonnull
        private static EmptyIngredient read(class_9129 buffer) {
            String item = buffer.method_19772();
            String tagId = buffer.method_19772();
            if (!item.isEmpty()) {
                return new EmptyIngredient(class_7923.field_41178.method_10223(class_2960.method_60654(item)));
            }
            if (!tagId.isEmpty()) {
                class_6862<class_1792> tag = class_6862.method_40092(class_7924.field_41197, class_2960.method_60654(tagId));
                return new EmptyIngredient(tag);
            }
            return new EmptyIngredient();
        }

        private static void write(@Nonnull class_9129 buffer, @Nonnull EmptyIngredient ingredient) {
            buffer.method_10814(ingredient.item != null ? Objects.requireNonNull(class_7923.field_41178.method_10221(ingredient.item)).toString() : "");
            buffer.method_10814(ingredient.tag != null ? ingredient.tag.comp_327().toString() : "");
        }
    }

}
